"""Useful functions that make easier develop other functions.
"""

import numpy as np
import pandas as pd


def _to_df(*args, **kwargs):
    res = None
    if args != ():
        res = pd.DataFrame([*args]).T

    if kwargs is not None:
        if res is not None:
            res = pd.concat(
                [res, pd.DataFrame.from_dict(kwargs, orient="columns")], axis=1
            )
        else:
            res = pd.DataFrame.from_dict(kwargs, orient="columns")
    return res


def _apply_to_df(func, df, x, weights, *args, **kwargs):
    """This function generalize main arguments as Series of a pd.Dataframe.
    Parameters
    ---------
    func : function
        Function to convert his arguments in Series of an Dataframe.
    df : pandas.Dataframe
        DataFrame whats contains the Series `x_name` and `w_name`
    x_name : str
        Name of the column in `df`
    weights_name : str
        Name of the column in `df
    Returns
    -------
    return : func return
        It's depends of func output type
    """
    return func(df[x], df[weights], *args, **kwargs)


def not_empty_weights(weights, as_of):
    if not (weights is None):
        w = normalize(weights.copy())
    else:
        w = np.ones(len(as_of))
    return w


def not_null_condition(income, weights):

    if np.any(income <= 0):
        mask = income > 0
        income = income[mask]
        if weights is not None:
            weights = weights[mask]

    return income, weights


def _sort_values(values, partner):
    idx_sort = np.argsort(values)
    values = values[idx_sort]
    partner = partner[idx_sort]
    return values, partner


def _clean_nans_values(this, pair):
    if np.any(np.isnan(this)):
        idx = ~np.isnan(this)
        this = this[idx]
        pair = pair[idx]
    return this, pair


def normalize(this):
    return this / np.sum(this)


def extract_values(data, variable, weights):
    if data is not None:
        variable = data.loc[:, variable].values
        weights = not_empty_weights(
            data.loc[:, weights].values, as_of=variable
        )
    return variable, weights


def repeat_data_from_weighted(x, w):
    if isinstance(w[0], float):
        raise NotImplementedError

    repeated_x = np.array([])
    repeated_w = np.array([])

    for xi, wi in zip(x, w):
        repeated_x = np.append(repeated_x, np.repeat(xi, wi))
        repeated_w = np.append(repeated_w, np.ones(wi))

    return repeated_x, repeated_w


def generate_data_to_test(n_sample_range=(20, 100)):
    N_sample = np.random.randint(*n_sample_range)
    weighted_x = np.random.randint(0, 1000, N_sample)
    weights = np.random.randint(1, 9, N_sample)
    return weighted_x, weights



import numpy as np

def ede(a, epsilon = None, kappa = None, weights = None):
    """
    Compute the Kolm-Pollak Equally-Distributed Equivalent (EDE).
    The Kolm-Pollak EDE and Index are suitable for distributions of desirable and
    undesirable quantities. That is, a desirable quantity (like income) is where
    having more of the quantity is desirable; compared with an undesirable
    quantity like health risk, where less is better.

    Parameters
    ----------
    a : array_like
        1-D array containing the values of the distribution.
    epsilon : float
        The inequality aversion parameter from the Atkinson formulae.
            If epsilon > 0 then the quantity is desirable (more is better).
    kappa : float
        The inequality aversion parameter from the Kolm-Pollak formulae.
            If kappa > 0 then the quantity is desirable (more is better).
    weights : array_like, optional
        1-D array of integer weights associated with the values in `a`. Each value in
        `a` contributes to the average according to its associated weight.
        If `weights=None`, then all data in `a` are assumed to have a
        weight equal to one.

    Returns
    -------
    value : float
        Returns the Kolm-Pollak Equally-Distributed Equivalent of the distribution provided.
    """
    a = np.asanyarray(a)
    if kappa is None:
        if epsilon is None:
            raise TypeError("you must provide either a epsilon or kappa aversion parameter")
        kappa = calc_kappa(a, epsilon, weights)
    if weights is None:
        ede_sum = np.exp(a*-kappa).sum()
        N = len(a)
    else:
        ede_sum = np.multiply(np.exp(a*-kappa), weights).sum()
        N = sum(weights)
    return(-1 / kappa) * np.log(ede_sum / N)


def index(a, epsilon = None, kappa = None, weights = None):
    """
    Compute the Kolm-Pollak Index.
    The Kolm-Pollak EDE and Index are suitable for distributions of desirable and
    undesirable quantities. That is, a desirable quantity (like income) is where
    having more of the quantity is desirable; compared with an undesirable
    quantity like health risk, where less is better.

    Parameters
    ----------
    a : array_like
        1-D array containing the values of the distribution.
    epsilon : float
        The inequality aversion parameter from the Atkinson formulae.
            If epsilon > 0 then the quantity is desirable (more is better).
    kappa : float
        The inequality aversion parameter from the Kolm-Pollak formulae.
            If kappa > 0 then the quantity is desirable (more is better).
    weights : array_like, optional
        1-D array of integer weights associated with the values in `a`. Each value in
        `a` contributes to the average according to its associated weight.
        If `weights=None`, then all data in `a` are assumed to have a
        weight equal to one.

    Returns
    -------
    value : float
        Returns the Kolm-Pollak Index of the distribution provided.
    """
    if weights is None:
        x_mean = np.mean(a)
    else:
        x_mean = np.average(a, weights = weights)

    return ede(a, epsilon = epsilon, kappa = kappa, weights = weights) - x_mean


def calc_kappa(a, epsilon, weights = None):
    """
    Converts the inequality aversion parameter used in Atkinson's formulae (epsilon)
    into the form for the Kolm-Pollak formulae (kappa).
    If epsilon > 0 then the quantity is desirable (more is better).

    Parameters
    ----------
    a : array_like
        1-D array containing the values of the distribution.
    epsilon : float
        The inequality aversion parameter from the Atkinson formulae.
            If epsilon > 0 then the quantity is desirable (more is better).
    weights : array_like, optional
        1-D array of integer weights associated with the values in `a`. Each value in
        `a` contributes to the average according to its associated weight.
        If `weights=None`, then all data in `a` are assumed to have a
        weight equal to one.

    Returns
    -------
    value : float
        Returns the inequality aversion parameter for the Kolm-Pollak formulae
    """
    if weights is None:
        x_sum = sum(a)
        x_sq_sum = (np.array(a)**2).sum()
    else:
        x_sum = np.multiply(a, weights).sum()
        x_sq_sum = np.multiply(a**2, weights).sum()
    return(epsilon * (x_sum / x_sq_sum))



import numpy as np

def index(a, weights = None):
    """
    Compute the Gini Coefficient.
    Thanks Gaëtan de Menten: https://stackoverflow.com/a/49571213/5890574

    Parameters
    ----------
    a : array_like
        1-D array containing the values of the distribution.
    weights : array_like, optional
        1-D array of integer weights associated with the values in `a`. Each value in
        `a` contributes to the average according to its associated weight.
        If `weights=None`, then all data in `a` are assumed to have a
        weight equal to one.

    Returns
    -------
    gini : float
        Returns the Gini Coefficient of the distribution provided.
    """
    a = np.asarray(a)
    if weights is not None:
        weights = np.asarray(weights)
        sorted_indices = np.argsort(a)
        sorted_x = a[sorted_indices]
        sorted_w = weights[sorted_indices]
        # Force float dtype to avoid overflows
        cumw = np.cumsum(sorted_w, dtype=float)
        cumxw = np.cumsum(sorted_x * sorted_w, dtype=float)
        return (np.sum(cumxw[1:] * cumw[:-1] - cumxw[:-1] * cumw[1:]) /
                (cumxw[-1] * cumw[-1]))
    else:
        sorted_x = np.sort(a)
        n = len(a)
        cumx = np.cumsum(sorted_x, dtype=float)
        # The above formula, with all weights equal to 1 simplifies to:
        return (n + 1 - 2 * np.sum(cumx) / cumx[-1]) / n



"""Low level desciptive statistics.

References
----------

1. http://people.ds.cam.ac.uk/fanf2/hermes/doc/antiforgery/stats.pdf

2. https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_variance

3. https://en.wikipedia.org/wiki/Algorithms%5Ffor%5Fcalculating%5Fvariance
   #Weighted_incremental_algorithm

4. https://www.thinkbrg.com/media/publication/720_McCrary_ImplementingAlgorithms_
   Whitepaper_20151119_WEB.pdf
"""

import numpy as np
from numba import guvectorize

from . import utils


def c_moment(variable=None, weights=None, order=2, param=None, ddof=0):
    """Calculate the central moment of `x` with respect to `param` of order `n`,
    given the weights `w`.

    Parameters
    ----------
    variable : 1d-array
        Variable
    weights : 1d-array
        Weights
    order : int, optional
        Moment order, 2 by default (variance)
    param : int or array, optional
        Parameter for which the moment is calculated, the default is None,
        implies use the mean.
    ddof : int, optional
        Degree of freedom, zero by default.

    Returns
    -------
    central_moment : float

    Notes
    -----
    - The cmoment of order 1 is 0
    - The cmoment of order 2 is the variance.

    Source : https://en.wikipedia.org/wiki/Moment_(mathematics)

    TODO
    ----
    Implement : https://en.wikipedia.org/wiki/L-moment#cite_note-wang:96-6

    """
    # return np.sum((x-c)^n*counts) / np.sum(counts)
    variable = variable.copy()
    weights = utils.not_empty_weights(weights, as_of=variable)

    if param is None:
        param = mean(variable=variable, weights=weights)
    elif not isinstance(param, (np.ndarray, int, float)):
        raise NotImplementedError

    return np.sum((variable - param) ** order * weights) / (
        np.sum(weights) - ddof
    )


def percentile(variable, weights, percentile=50, interpolation="lower"):
    """Calculate the percentile.

    Parameters
    ----------
    variable : str or array
    weights :  str or array
    percentile : int or list
        Percentile level, if pass 50 we get the median.
    Returns
    -------
    percentile : float
    """

    sorted_idx = np.argsort(variable)
    cum_weights = np.cumsum(weights[sorted_idx])
    lower_percentile_idx = np.searchsorted(
        cum_weights, (percentile / 100.0) * cum_weights[-1]
    )

    if interpolation is "midpoint":
        res = np.interp(
            lower_percentile_idx + 0.5,
            np.arange(len(variable)),
            variable[sorted_idx],
        )
    elif interpolation is "lower":
        res = variable[sorted_idx[lower_percentile_idx]]
    elif interpolation is "higher":
        res = variable[sorted_idx[lower_percentile_idx + 1]]
    else:
        raise NotImplementedError

    return res


def std_moment(variable=None, weights=None, param=None, order=3, ddof=0):
    """Calculate the standarized moment.

    Calculate the standarized moment of order `c` for the variable` x` with
    respect to `c`.

    Parameters
    ----------
    variable : 1d-array
       Random Variable
    weights : 1d-array, optional
       Weights or probability
    order : int, optional
       Order of Moment, three by default
    param : int or float or array, optional
       Central trend, default is the mean.
    ddof : int, optional
        Degree of freedom.

    Returns
    -------
    std_moment : float
       Returns the standardized `n` order moment.

    References
    ----------
    - https://en.wikipedia.org/wiki/Moment_(mathematics)#Significance_of_the_moments
    - https://en.wikipedia.org/wiki/Standardized_moment

    TODO
    ----
    It is the general case of the raw and central moments. Review
    implementation.

    """
    if param is None:
        param = mean(variable=variable, weights=weights)

    res = c_moment(
        variable=variable, weights=weights, order=order, param=param, ddof=ddof
    )
    res /= var(variable=variable, weights=weights, ddof=ddof) ** (order / 2)
    return res


def mean(variable=None, weights=None):
    """Calculate the mean of `variable` given `weights`.

    Parameters
    ----------
    variable : array-like or str
        Variable on which the mean is estimated.
    weights : array-like or str
        Weights of the `x` variable.

    Returns
    -------
    mean : array-like or float
    """
    # if pass a DataFrame separate variables.
    variable = variable.copy()
    weights = utils.not_empty_weights(weights, as_of=variable)
    variable, weights = utils._clean_nans_values(variable, weights)
    return np.average(a=variable, weights=weights, axis=0)


def var(variable=None, weights=None, ddof=0):
    """Calculate the population variance of ``variable`` given `weights`.

    Parameters
    ----------
    variable : 1d-array or pd.Series or pd.DataFrame
        Variable on which the quasivariation is estimated
    weights : 1d-array or pd.Series or pd.DataFrame
        Weights of the `variable`.

    Returns
    -------
    variance : 1d-array or pd.Series or float
        Estimation of quasivariance of `variable`

    References
    ---------
    Moment (mathematics). (2017, May 6). In Wikipedia, The Free Encyclopedia.
    Retrieved 14:40, May 15, 2017, from
    https://en.wikipedia.org/w/index.php?title=Moment_(mathematics)&oldid=778996402

    Notes
    -----
    If stratificated sample must pass with groupby each strata.
    """
    return c_moment(variable=variable, weights=weights, order=2, ddof=ddof)


def coef_variation(variable=None, weights=None):
    """Calculate the coefficient of variation.

     Calculate the coefficient of variation of a `variable` given weights.
    The coefficient of variation is the square root of the variance of the
    incomes divided by the mean income. It has the advantages of being
    mathematically tractable and is subgroup decomposable, but is not bounded
    from above.

    Parameters
    ----------
    variable : array-like or str
    weights : array-like or str

    Returns
    -------
    coefficient_variation : float

    References
    ----------
    Coefficient of variation. (2017, May 5). In Wikipedia, The Free Encyclopedia.
    Retrieved 15:03, May 15, 2017, from
    https://en.wikipedia.org/w/index.php?title=Coefficient_of_variation&oldid=778842331
    """
    # todo complete docstring
    return var(variable=variable, weights=weights) ** 0.5 / abs(
        mean(variable=variable, weights=weights)
    )


def kurt(variable=None, weights=None):
    """Calculate the asymmetry coefficient

    Parameters
    ---------
    variable : 1d-array
    weights : 1d-array

    Returns
    -------
    kurt : float
        Kurtosis coefficient.

    References
    ---------
    Moment (mathematics). (2017, May 6). In Wikipedia, The Free Encyclopedia.
    Retrieved 14:40, May 15, 2017, from
    https://en.wikipedia.org/w/index.php?title=Moment_(mathematics)&oldid=778996402

    Notes
    -----
    It is an alias of the standardized fourth-order moment.

    """
    return std_moment(variable=variable, weights=weights, order=4)


def skew(variable=None, weights=None):
    """Returns the asymmetry coefficient of a sample.

    Parameters
    ---------
    variable : array-like, str
    weights : array-like, str

    Returns
    -------
    skew : float

    References
    ---------
    Moment (mathematics). (2017, May 6). In Wikipedia, The Free Encyclopedia.
    Retrieved 14:40, May 15, 2017, from
    https://en.wikipedia.org/w/index.php?title=Moment_(mathematics)&oldid=778996402

    Notes
    -----
    It is an alias of the standardized third-order moment.

    """
    return std_moment(variable=variable, weights=weights, order=3)


@guvectorize(
    "float64[:], float64[:], int64, float64[:]",
    "(n),(n),()->()",
    nopython=True,
    cache=True,
)
def wvar(x, w, kind, out):
    """Calculate weighted variance of X.

    Calculates the weighted variance of x according to a kind of weights.

    Parameters
    ----------
    x : np.ndarray
        Main variable.
    w : np.ndarray
        Weigths.
    kind : int
        Has three modes to calculate de variance, you can control that with this
        argument, the values and the output are the next:
        1 : population variance
        2 : sample frequency variance
        3 : sample reliability variance.
    out : np.ndarray

    Returns
    -------
    weighted_variance : float

    References
    ----------
    https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Weighted_incremental_algorithm
    """
    wSum = wSum2 = mean = S = 0

    for i in range(len(x)):  # Alternatively "for x, w in zip(data, weights):"
        wSum = wSum + w[i]
        wSum2 = wSum2 + w[i] * w[i]
        meanOld = mean
        mean = meanOld + (w[i] / wSum) * (x[i] - meanOld)
        S = S + w[i] * (x[i] - meanOld) * (x[i] - mean)

    if kind == 1:
        # population_variance
        out[0] = S / wSum
    elif kind == 2:
        # Bessel's correction for weighted samples
        # Frequency weights
        # sample_frequency_variance
        out[0] = S / (wSum - 1)
    elif kind == 3:
        # Reliability weights
        # sample_reliability_variance
        out[0] = S / (wSum - wSum2 / wSum)


@guvectorize(
    "float64[:], float64[:], float64[:], int64, float64[:]",
    "(n),(n),(n),()->()",
    nopython=True,
    cache=True,
)
def wcov(x, y, w, kind, out):
    """Compute weighted covariance between x and y.

    Compute the weighted covariance between two variables, we can chose which
    kind of covariance returns.

    Parameters
    ----------
    x : np.array
        Main variable.
    y : np.array
        Second variable.
    w : np.array
        Weights.
    kind : int
        Kind of weighted covariance is returned:
            1 : population variance
            2 : sample frequency variance
            3 : sample reliability variance.
    out : np.array

    Returns
    -------
    weighted_covariance = float

    References
    ----------
    https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Online
    """
    meanx = meany = 0
    wsum = wsum2 = 0
    C = 0
    for i in range(len(x)):
        wsum += w[i]
        wsum2 += w[i] * w[i]
        dx = x[i] - meanx
        meanx += (w[i] / wsum) * dx
        meany += (w[i] / wsum) * (y[i] - meany)
        C += w[i] * dx * (y[i] - meany)

    if kind == 1:
        # population_covar
        out[0] = C / wsum
    elif kind == 1:
        # Bessel's correction for sample variance
        # Frequency weights
        # sample_frequency_covar
        out[0] = C / (wsum - 1)
    elif kind == 1:
        # Reliability weights
        # sample_reliability_covar
        out[0] = C / (wsum - wsum2 / wsum)


@guvectorize(
    "float64[:], float64[:], float64[:]",
    "(n),(n)->()",
    nopython=True,
    cache=True,
)
def online_kurtosis(x, w, out):
    n = mean = M2 = M3 = M4 = 0

    for i in range(len(x)):
        n1 = w[i]
        n = n + w[i]
        delta = x[i] - mean
        delta_n = delta / n
        delta_n2 = delta_n * delta_n
        term1 = delta * delta_n * n1
        mean = mean + w[i] * delta_n / n
        M4 = (
            M4
            + term1 * delta_n2 * (n * n - 3 * n + 3)
            + 6 * delta_n2 * M2
            - 4 * delta_n * M3
        )
        M3 = M3 + term1 * delta_n * (n - 2) - 3 * delta_n * M2
        M2 = M2 + term1

    kurtosis = (n * M4) / (M2 * M2) - 3


@guvectorize(
    "float64[:], float64[:], int64, float64[:]",
    "(n),(n),()->()",
    nopython=True,
    cache=True,
)
def Mk(x, w, k, out):
    w_sum = wx_sum = 0

    for i in range(len(x)):
        wx_sum += w[i] * (x[i] ** k)
        w_sum += w[i]

    out[0] = wx_sum / w_sum



import numpy as np

def ede(a, epsilon = 0.5, weights = None):
    """
    Compute the Atkinson Equally-Distributed Equivalent.
    The Atkinson EDE and Index are only suitable for distributions of desirable
    quantities (where having more of the quantity is desirable), e.g., income.

    Parameters
    ----------
    a : array_like
        1-D array containing the values of the distribution.
    epsilon : float
        The inequality aversion parameter. epsilon > 0.
    weights : array_like, optional
        1-D array of integer weights associated with the values in `a`. Each value in
        `a` contributes to the average according to its associated weight.
        If `weights=None`, then all data in `a` are assumed to have a
        weight equal to one.

    Returns
    -------
    value : float
        Returns the Atkinson EDE of the distribution provided.
    """
    sum_atk = 0
    if not weights:
        N = len(a)
    else:
        N = sum(weights)
    x_mean = np.average(a, weights = weights)
    count = 0
    for i in a:
        if not weights:
            sum_atk += i**(1 - epsilon)
        else:
            sum_atk += (i**(1 - epsilon)) * weights[count]
        count += 1
    ede = (sum_atk / N)**(1 / (1 - epsilon))
    return(ede)


def index(a, epsilon = 0.5, weights = None):
    """
    Compute the Atkinson Index.
    The Atkinson EDE and Index are only suitable for distributions of desirable
    quantities (where having more of the quantity is desirable), e.g., income.

    Parameters
    ----------
    a : array_like
        1-D array containing the values of the distribution.
    epsilon : float
        The inequality aversion parameter. epsilon > 0.
    weights : array_like, optional
        1-D array of integer weights associated with the values in `a`. Each value in
        `a` contributes to the average according to its associated weight.
        If `weights=None`, then all data in `a` are assumed to have a
        weight equal to one.

    Returns
    -------
    value : float
        Returns the Atkinson Index of the distribution provided.
    """
    ede_atk = ede(a, epsilon, weights)
    x_mean = np.average(a, weights = weights)

    return(1 - (ede_atk / x_mean))



#!/usr/bin/env python

"""Analysis of inequality.

This package provide an easy way to realize a quantitative analysis of
grouped, also make easy work with stratified data, in this module you can
find statistics and grouped indicators to this task.

Todo
----
- Rethinking this module as Class.
- https://en.wikipedia.org/wiki/Income_inequality_metrics

"""
import numpy as np
import pandas as pd
from .statistics import mean
from . import utils


def concentration(income, weights=None, data=None, sort=True):
    """This function calculate the concentration index, according to the
    notation used in [Jenkins1988]_ you can calculate the:
    C_x = 2 / x · cov(x, F_x)
    if x = g(x) then C_x becomes C_y
    when there are taxes:

    y = g(x) = x - t(x)

    Parameters
    ----------
    income : array-like
    weights : array-like
    data : pandas.DataFrame
    sort : bool

    Returns
    -------
    concentration : array-like

    References
    ----------
    Jenkins, S. (1988). Calculating income distribution indices
    from micro-data. National Tax Journal. http://doi.org/10.2307/41788716
    """
    # TODO complete docstring

    # check if DataFrame is passed, if yes then extract variables else make a copy
    income, weights = utils.extract_values(data, income, weights)
    if weights is None:
        weights = utils.not_empty_weights(weights, as_of=income)
    # if sort is true then sort the variables.
    if sort:
        income, weights = utils._sort_values(income, weights)
    # main calc
    f_x = utils.normalize(weights)
    F_x = f_x.cumsum()
    mu = np.sum(income * f_x)
    cov = np.cov(income, F_x, rowvar=False, aweights=f_x)[0, 1]
    return 2 * cov / mu


def lorenz(income, weights=None, data=None):
    """In economics, the Lorenz curve is a graphical representation of the
    distribution of income or of wealth. It was developed by Max O. Lorenz in
    1905 for representing grouped of the wealth distribution. This function
    compute the lorenz curve and returns a DF with two columns of axis x and y.

    Parameters
    ----------
    data : pandas.DataFrame
        A pandas.DataFrame that contains data.
    income : str or 1d-array, optional
        Population or wights, if a DataFrame is passed then `income` should be a
        name of the column of DataFrame, else can pass a pandas.Series or array.
    weights : str or 1d-array
        Income, monetary variable, if a DataFrame is passed then `y`is a name
        of the series on this DataFrame, however, you can pass a pd.Series or
        np.array.

    Returns
    -------
    lorenz : pandas.Dataframe
        Lorenz distribution in a Dataframe with two columns, labeled x and y,
        that corresponds to plots axis.

    References
    ----------
    Lorenz curve. (2017, February 11). In Wikipedia, The Free Encyclopedia.
    Retrieved 14:34, May 15, 2017, from
    https://en.wikipedia.org/w/index.php?title=Lorenz_curve&oldid=764853675
    """

    if data is not None:
        income, weights = utils.extract_values(data, income, weights)

    total_income = income * weights
    idx_sort = np.argsort(income)

    weights = weights[idx_sort].cumsum() / weights.sum()
    weights = weights.reshape(len(weights), 1)

    total_income = total_income[idx_sort].cumsum() / total_income.sum()
    total_income = total_income.reshape(len(total_income), 1)

    # to pandas
    data = np.hstack([weights, total_income])
    columns = ["Equality", "Income"]
    index = pd.Index(weights.round(3).squeeze())
    res = pd.DataFrame(data=data, columns=columns, index=index)
    res.index.name = "x"

    return res


def gini(income, weights=None, data=None, sort=True):
    """The Gini coefficient (sometimes expressed as a Gini ratio or a
    normalized Gini index) is a measure of statistical dispersion intended to
    represent the income or wealth distribution of a nation's residents, and is
    the most commonly used measure of grouped. It was developed by Corrado
    Gini.
    The Gini coefficient measures the grouped among values of a frequency
    distribution (for example, levels of income). A Gini coefficient of zero
    expresses perfect equality, where all values are the same (for example,
    where everyone has the same income). A Gini coefficient of 1 (or 100%)
    expresses maximal grouped among values (e.g., for a large number of
    people, where only one person has all the income or consumption, and all
    others have none, the Gini coefficient will be very nearly one).

    Parameters
    ---------
    data : pandas.DataFrame
        DataFrame that contains the data.
    income : str or np.array, optional
        Name of the monetary variable `x` in` df`
    weights : str or np.array, optional
        Name of the series containing the weights `x` in` df`
    sorted : bool, optional
        If the DataFrame is previously ordered by the variable `x`, it's must
        pass True, but False by default.

    Returns
    -------
    gini : float
        Gini Index Value.

    Notes
    -----
    The calculation is done following (discrete probability distribution):
    G = 1 - [∑_i^n f(y_i)·(S_{i-1} + S_i)]
    where:
    - y_i = Income
    - S_i = ∑_{j=1}^i y_i · f(y_i)

    Reference
    ---------
    - Gini coefficient. (2017, May 8). In Wikipedia, The Free Encyclopedia.
      Retrieved 14:30, May 15, 2017, from
      https://en.wikipedia.org/w/index.php?title=Gini_coefficient&oldid=779424616

    - Jenkins, S. (1988). Calculating income distribution indices
    from micro-data. National Tax Journal. http://doi.org/10.2307/41788716

    TODO
    ----
    - Implement statistical deviation calculation, VAR (GINI)

    """
    return concentration(data=data, income=income, weights=weights, sort=sort)


def atkinson(income, weights=None, data=None, e=0.5):
    """More precisely labelled a family of income grouped measures, the
    theoretical range of Atkinson values is 0 to 1, with 0 being a state of
    equal distribution.
    An intuitive interpretation of this index is possible: Atkinson values can
    be used to calculate the proportion of total income that would be required
    to achieve an equal level of social welfare as at present if incomes were
    perfectly distributed.

    For example, an Atkinson index value of 0.20 suggests
    that we could achieve the same level of social welfare with only
    1 – 0.20 = 80% of income. The theoretical range of Atkinson values is 0 to 1,
    with 0 being a state of equal distribution.

    Parameters
    ---------
    income : array or str
        If `data` is none `income` must be an 1D-array, when `data` is a
        pd.DataFrame, you must pass the name of income variable as string.
    weights : array or str, optional
        If `data` is none `weights` must be an 1D-array, when `data` is a
        pd.DataFrame, you must pass the name of weights variable as string.
    e : int, optional
        Epsilon parameter interpreted by atkinson index as grouped adversion,
        must be between 0 and 1.
    data : pd.DataFrame, optional
        data is a pd.DataFrame that contains the variables.

    Returns
    -------
    atkinson : float

    Reference
    ---------
    Atkinson index. (2017, March 12). In Wikipedia, The Free Encyclopedia.
    Retrieved 14:35, May 15, 2017, from
    https://en.wikipedia.org/w/index.php?title=Atkinson_index&oldid=769991852

    TODO
    ----
    - Implement: CALCULATING INCOME DISTRIBUTION INDICES FROM MICRO-DATA
      http://www.jstor.org/stable/41788716
    - The results has difference with stata, maybe have a bug.
    """
    if (income is None) and (data is None):
        raise ValueError("Must pass at least one of both `income` or `df`")

    income, weights = utils.extract_values(data, income, weights)
    weights = utils.not_empty_weights(weights, income)

    # not-null condition
    income, weights = utils.not_null_condition(income, weights)

    # not-empty condition
    if len(income) == 0:
        return 0

    # auxiliar variables: mean and distribution
    mu = mean(variable=income, weights=weights)
    f_i = weights / sum(weights)  # density function

    # main calc
    if e == 1:
        atkinson = 1 - np.power(
            np.e, np.sum(f_i * np.log(income) - np.log(mu))
        )
    elif (0 <= e) or (e < 1):
        atkinson = 1 - np.power(
            np.sum(f_i * np.power(income / mu, 1 - e)), 1 / (1 - e)
        )
    else:
        assert (e < 0) or (e > 1), "Not valid e value,  0 ≤ e ≤ 1"
        atkinson = None
    return atkinson


def kakwani(tax, income_pre_tax, weights=None, data=None):
    """The Kakwani (1977) index of tax progressivity is defined as twice the
    area between the concentration curves for taxes and pre-tax income,
    or equivalently, the concentration index for t(x) minus the Gini index for
    x, i.e.

    K = C(t) - G(x)
      = (2/t) cov [t(x), F(x)] - (2/x) cov [x, F(x)].

    Parameters
    ----------
    data : pandas.DataFrame
        This variable is a DataFrame that contains all data required in
        columns.
    tax_variable : array-like or str
        This variable represent tax payment of person, if pass array-like
        then data must be None, else you pass str-name column in `data`.
    income_pre_tax : array-like or str
        This variable represent income of person, if pass array-like
        then data must be None, else you pass str-name column in `data`.
    weights : array-like or str
        This variable represent weights of each person, if pass array-like
        then data must be None, else you pass str-name column in `data`.

    Returns
    -------
    kakwani : float

    References
    ----------
    Jenkins, S. (1988). Calculating income distribution indices from micro-data.
    National Tax Journal. http://doi.org/10.2307/41788716
    """
    # main calc
    c_t = concentration(data=data, income=tax, weights=weights, sort=True)
    g_y = concentration(
        data=data, income=income_pre_tax, weights=weights, sort=True
    )
    return c_t - g_y


def reynolds_smolensky(
    income_pre_tax, income_post_tax, weights=None, data=None
):
    """The Reynolds-Smolensky (1977) index of the redistributive effect of
    taxes, which can also be interpreted as an index of progressivity
    (Lambert 1985), is defined as:

    L = Gx - Gy
      = [2/x]cov[x,F(x)] - [2/ybar] cov [y, F(y)].

    Parameters
    ----------
    data : pandas.DataFrame
        This variable is a DataFrame that contains all data required in it's
        columns.
    income_pre_tax : array-like or str
        This variable represent tax payment of person, if pass array-like
        then data must be None, else you pass str-name column in `data`.
    income_post_tax : array-like or str
        This variable represent income of person, if pass array-like
        then data must be None, else you pass str-name column in `data`.
    weights : array-like or str
        This variable represent weights of each person, if pass array-like
        then data must be None, else you pass str-name column in `data`.

    Returns
    -------
    reynolds_smolensky : float

    References
    ----------
    Jenkins, S. (1988). Calculating income distribution indices from micro-data.
    National Tax Journal. http://doi.org/10.2307/41788716
    """
    g_y = concentration(data=data, income=income_post_tax, weights=weights)
    g_x = concentration(data=data, income=income_pre_tax, weights=weights)
    return g_x - g_y


def theil(income, weights=None, data=None):
    """The Theil index is a statistic primarily used to measure economic
    grouped and other economic phenomena. It is a special case of the
    generalized entropy index. It can be viewed as a measure of redundancy,
    lack of diversity, isolation, segregation, grouped, non-randomness, and
    compressibility. It was proposed by econometrician Henri Theil.

    Parameters
    ----------
    data : pandas.DataFrame
        This variable is a DataFrame that contains all data required in it's
        columns.
    income : array-like or str
        This variable represent tax payment of person, if pass array-like
        then data must be None, else you pass str-name column in `data`.
    weights : array-like or str
        This variable represent weights of each person, if pass array-like
        then data must be None, else you pass str-name column in `data`.

    Returns
    -------
    theil : float

    References
    ----------
    Theil index. (2016, December 17). In Wikipedia, The Free Encyclopedia.
    Retrieved 14:17, May 15, 2017, from
    https://en.wikipedia.org/w/index.php?title=Theil_index&oldid=755407818

    """
    if data is not None:
        income, weights = utils.extract_values(data, income, weights)
    else:
        income = income.copy()
        weights = weights.copy()
    income, weights = utils.not_null_condition(income, weights)

    # variables needed
    mu = mean(variable=income, weights=weights)
    f_i = utils.normalize(weights)
    # main calc
    theil = np.sum((f_i * income / mu) * np.log(income / mu))
    return theil


def avg_tax_rate(total_tax, total_base, weights=None, data=None):
    """This function compute the average tax rate given a base income and a
    total tax.

    Parameters
    ----------
    total_base : str or numpy.array
    total_tax : str or numpy.array
    data : pd.DataFrame

    Returns
    -------
    avg_tax_rate : float or pd.Series
        Is the ratio between mean the tax income and base of income.

    Reference
    ---------
    Panel de declarantes de IRPF 1999-2007: Metodología, estructura y variables.
    (2011). Panel de declarantes de IRPF 1999-2007:
    Metodología, estructura y variables. Documentos.
    """
    if isinstance(total_base, (np.ndarray)):
        n_cols = total_base.shape[1]
    elif isinstance(total_base, (list)):
        n_cols = len(total_base)
    elif isinstance(total_base, (str)):
        n_cols = 1
    else:
        n_cols = total_base.shape[1]

    numerator = mean(data=data, variable=total_tax, weights=weights)
    denominator = mean(data=data, variable=total_base, weights=weights)
    # main calc
    res = numerator / denominator

    if data is not None:
        base_name = total_base
        tax_name = total_tax
    else:
        base_name = ["base" % i for i in range(n_cols)]
        tax_name = ["tax_%s" % i for i in range(n_cols)]

    names = ["_".join([t, b]) for t, b in zip(tax_name, base_name)]
    res = pd.Series(res, index=names)
    return res



import setuptools

#Get long description from readme file
with open("README.md", 'r') as readme_file:
    long_description = readme_file.read()

setuptools.setup(
    name="inequalipy",
    version="1.0.2",
    author="Tom Logan, Mitchell Anderson",
    author_email="tom.logan@canterbury.ac.nz",
    description="Inequalipy contains functions for the Kolm-Pollak, Atkinson and Gini Index aproaches to calculating inequality of a distribution",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/urutau-nz/inequalipy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy'
    ],
    python_requires='>=3',
)



"""Descriptive statistics.

This module contains main descriptive statistics like: mean, variance, etc.

"""

import pandas as pd
from . import _statistics as stat
from . import utils


def c_moment(
    variable=None, weights=None, data=None, order=2, param=None, ddof=0
):
    """Calculate the central moment of `x` with respect to `param` of order `n`,
    given the weights `w`.

    Parameters
    ----------
    variable : 1d-array
        Variable
    weights : 1d-array
        Weights
    data : pandas.DataFrame
        Contains all variables needed.
    order : int, optional
        Moment order, 2 by default (variance)
    param : int or array, optional
        Parameter for which the moment is calculated, the default is None,
        implies use the mean.
    ddof : int, optional
        Degree of freedom, zero by default.

    Returns
    -------
    central_moment : float

    Notes
    -----
    - The cmoment of order 1 is 0
    - The cmoment of order 2 is the variance.

    Source : https://en.wikipedia.org/wiki/Moment_(mathematics)

    TODO
    ----
    Implement: https://en.wikipedia.org/wiki/L-moment#cite_note-wang:96-6

    """
    variable, weights = utils.extract_values(data, variable, weights)
    return stat.c_moment(variable, weights, order, param, ddof)


def percentile(
    variable=None, weights=None, data=None, p=50, interpolate="lower"
):
    """Calculate the value of a quantile given a variable and his weights.

    Parameters
    ----------
    variable : str or array
    weights :  str or array
    data : pd.DataFrame, optional
        pd.DataFrame that contains all variables needed.
    q : float
        Quantile level, if pass 0.5 means median.
    interpolate : bool

    Returns
    -------
    percentile : float or pd.Series

    """
    variable, weights = utils.extract_values(data, variable, weights)
    return stat.percentile(variable, weights, p, interpolate)


def std_moment(
    variable=None, weights=None, data=None, param=None, order=3, ddof=0
):
    """Calculate the standardized moment of order `c` for the variable` x` with
    respect to `c`.

    Parameters
    ----------
    variable : 1d-array
       Random Variable
    weights : 1d-array, optional
       Weights or probability
    data : pd.DataFrame, optional
        pd.DataFrame that contains all variables needed.
    order : int, optional
       Order of Moment, three by default
    param : int or float or array, optional
       Central trend, default is the mean.
    ddof : int, optional
        Degree of freedom.

    Returns
    -------
    std_moment : float
       Returns the standardized `n` order moment.

    References
    ----------
    - https://en.wikipedia.org/wiki/Moment_(mathematics)#Significance_of_the_moments
    - https://en.wikipedia.org/wiki/Standardized_moment

    TODO
    ----
    It is the general case of the raw and central moments. Review
    implementation.

    """
    variable, weights = utils.extract_values(data, variable, weights)
    return stat.std_moment(data, variable, weights, param, order, ddof)


def mean(variable=None, weights=None, data=None):
    """Calculate the mean of `variable` given `weights`.

    Parameters
    ----------
    variable : array-like or str
        Variable on which the mean is estimated.
    weights : array-like or str
        Weights of the `x` variable.
    data : pandas.DataFrame
        Is possible pass a DataFrame with variable and weights, then you must
        pass as `variable` and `weights` the column name stored in `data`.

    Returns
    -------
    mean : array-like or float
    """
    # if pass a DataFrame separate variables.
    if data is not None:
        variable, weights = utils.extract_values(data, variable, weights)
    return stat.mean(variable, utils.not_empty_weights(weights, variable))


def density(variable=None, weights=None, groups=None, data=None):
    """Density in percentage.

    Calculates density in percentage. This make division of variable inferring
    width in groups as max - min.

    Parameters
    ----------
    variable : numpy.array or pandas.DataFrame
        Main variable.
    weights : numpy.array or pandas.DataFrame
        Weights of main variable.
    groups : numpy.array or pandas.DataFrame
        Label that show which group each element belongs to.
    data : pd.DataFrame, optional
        Object that contains all variables needed.


    Returns
    -------
    density : array-like

    References
    ----------
    Histogram. (2017, May 9). In Wikipedia, The Free Encyclopedia. Retrieved
    14:47, May 15, 2017, from
    https://en.wikipedia.org/w/index.php?title=Histogram&oldid=779516918
    """
    variable, weights = utils.extract_values(data, variable, weights)
    if not (groups is None):
        groups = data[groups].values
    return stat.density(variable, weights, groups)


def var(variable=None, weights=None, data=None, ddof=0):
    """Calculate the variance.

    Calculate the population variance of `variable` given `weights`.

    Parameters
    ----------
    data : pd.DataFrame, optional
        pd.DataFrame that contains all variables needed.
    variable : 1d-array or pd.Series or pd.DataFrame
        Variable on which the quasivariation is estimated
    weights : 1d-array or pd.Series or pd.DataFrame
        Weights of the `variable`.
    data : pd.DataFrame
        Object that contains all variables needed.
    ddof : int
        Degree of freedom.

    Returns
    -------
    variance : 1d-array or pd.Series or float
        Estimation of quasivariance of `variable`

    References
    ---------
    Moment (mathematics). (2017, May 6). In Wikipedia, The Free Encyclopedia.
    Retrieved 14:40, May 15, 2017, from
    https://en.wikipedia.org/w/index.php?title=Moment_(mathematics)&oldid=
    778996402

    Notes
    -----
    If stratificated sample must pass with groupby each strata.
    """
    variable, weights = utils.extract_values(data, variable, weights)
    return stat.var(variable, weights, ddof)


def coef_variation(variable=None, weights=None, data=None):
    """Calculate the coefficient of variation.

    Calculate the coefficient of variation of a `variable` given weights.
    The coefficient of variation is the square root of the variance of the
    incomes divided by the mean income. It has the advantages of being
    mathematically tractable and is subgroup decomposable, but is not bounded
    from above.

    Parameters
    ----------
    variable : array-like or str
    weights : array-like or str
    data : pandas.DataFrame

    Returns
    -------
    coefficient_variation : float

    References
    ----------
    Coefficient of variation. (2017, May 5). In Wikipedia, The Free Encyclopedia.
    Retrieved 15:03, May 15, 2017, from https://en.wikipedia.org/w/index.php?
    title=Coefficient_of_variation&oldid=778842331
    """
    # TODO complete docstring
    variable, weights = utils.extract_values(data, variable, weights)
    return stat.coef_variation(variable, weights)


def kurt(variable=None, weights=None, data=None):
    """Calculate the Kurtosis coefficient.

    Parameters
    ---------
    variable : 1d-array
    weights : 1d-array
    data : pandas.DataFrame
        Object which stores ``variable`` and ``weights``.

    Returns
    -------
    kurt : float
        Kurtosis coefficient.

    References
    ---------
    Moment (mathematics). (2017, May 6). In Wikipedia, The Free Encyclopedia.
    Retrieved 14:40, May 15, 2017, from https://en.wikipedia.org/w/index.php?
    title=Moment_(mathematics)&oldid=778996402

    Notes
    -----
    It is an alias of the standardized fourth-order moment.
    """
    variable, weights = utils.extract_values(data, variable, weights)
    return stat.kurt(variable, weights)


def skew(variable=None, weights=None, data=None):
    """Returns the asymmetry coefficient of a sample.

    Parameters
    ---------
    data : pandas.DataFrame

    variable : array-like, str
    weights : array-like, str
    data : pandas.DataFrame
        Object which stores ``variable`` and ``weights``.

    Returns
    -------
    skew : float

    References
    ---------
    Moment (mathematics). (2017, May 6). In Wikipedia, The Free Encyclopedia.
    Retrieved 14:40, May 15, 2017, from
    https://en.wikipedia.org/w/index.php?title=Moment_(mathematics)&oldid=778996402

    Notes
    -----
    It is an alias of the standardized third-order moment.
    """
    variable, weights = utils.extract_values(data, variable, weights)
    return stat.skew(variable, weights)
